import numpy as np
from numba import jit

from neuralib.util.segments import SegmentLike
from neuralib.util.unstable import unstable

__all__ = ['running_mask1d',
           'jump_mask2d']


def running_mask1d(time: np.ndarray,
                   velocity: np.ndarray,
                   *,
                   threshold: float = 5,
                   merge_gap: float = 0.5,
                   minimal_time: float = 1) -> np.ndarray:
    """
    Running epoch mask for linear (1-dimension) locomotion.

    Refer to Zaremba et al., 2016. Nature Neuroscience.
    Selection based on criterion:

    - Forward locomotion
    - Velocity > threshold (default greater than 5 cm/s)
    - Merge the gap period (default 0.5 sec, if animal accidentally slow down)
    - Each segmented epoch need to longer than (default is 1 second)

    `Dimension parameters`:

        P = Number of position points


    :param time: 1D time array of the position. `Array[float, P]`
    :param velocity: 1D velocity array. `Array[float, P]`
    :param threshold: Velocity threshold in cm/s
    :param merge_gap: Merge gap in seconds (i.e., animal accidentally slow down in a short period of time)
    :param minimal_time: Epoch minimal time in seconds
    :return: Running epoch mask. `Array[bool, P]`
    """
    run = (velocity > threshold).astype(int)

    # epoch_list: [0*epoch0, 1*epoch1, 2*epoch2, ...]
    epoch = np.cumsum(abs(np.diff(run, prepend=run[0])))

    # measure even numbers of epoch index are gaps
    if velocity[0] >= threshold:
        epoch += 1

    # merge gap
    for i in range(2, np.max(epoch) + 1, 2):
        t = time[epoch == i]
        dt = np.max(t) - np.min(t)
        if dt < merge_gap:
            epoch[epoch == i] += 1
            epoch[epoch == i + 1] += 2

    # minimal time
    for i in range(1, np.max(epoch) + 1, 2):  # exclude 0 due to merge special case
        tt = time[epoch == i]
        if len(tt) > 0:  # i is not existed in ep_idx
            td = np.max(tt) - np.min(tt)
            if td < minimal_time:
                epoch[epoch == i] += 1

    return np.asarray(epoch % 2 != 0)


# noinspection PyTypeChecker
@unstable()
@jit(nopython=True)
def jump_mask2d(time: np.ndarray,
                xy: np.ndarray,
                jump_size: float,
                max_duration: float = 0.1) -> tuple[np.ndarray, SegmentLike]:
    """
    Find jump epoch in 2D xy locomotion

    :param time: A 1D numpy array of time points corresponding to each xy coordinate. `Array[bool, N]`
    :param xy: A 2D numpy array of x and y coordinates representing positions. `Array[float, [N, 2]]`
    :param jump_size: A float representing the minimum jump distance to be considered significant between consecutive points.
        Should be expressed in the same coordinates as xy
    :param max_duration: the maximum duration (in the same units as t) of a jump.
        If a jump lasts longer the duration, it will not be removed (as it may not actually be a jump)
    :return: A tuple containing:

        - A boolean numpy array indicating which points are part of a detected jump segment. `Array[bool, N]`

        - A list of segments, where each segment is represented by a list of two indices `[start_idx, end_idx]`.
    """
    valid = np.nonzero(~np.isnan(xy[:, 0]))[0]
    ret = np.zeros(xy.shape[0], dtype=np.bool_)
    seg = [[np.int64(x) for x in range(0)]]

    if len(valid) < 2:
        return ret, seg  # without detect

    found_jump = False
    last_idx = valid[0]

    for idx in valid[1:]:

        dt = time[idx] - time[last_idx]
        dxy = np.sqrt(np.sum((xy[idx] - xy[last_idx]) ** 2))

        if not found_jump and (dxy / (idx - last_idx)) > jump_size:
            found_jump = True

        if found_jump and (dxy <= jump_size or dt >= max_duration):

            if dt < max_duration:
                ret[last_idx:idx] = 1
                seg.append([last_idx, idx])

            found_jump = False

        if not found_jump:
            last_idx = idx

    return ret, seg[1:]



from __future__ import annotations

import numpy as np
from matplotlib import pyplot as plt

from neuralib.typing import PathLike

__all__ = ['plot_scatter_animation']


def plot_scatter_animation(x: np.ndarray,
                           y: np.ndarray,
                           t: np.ndarray | None = None, *,
                           step: int | None = None,
                           size: int = 10,
                           output: PathLike | None = None,
                           **kwargs) -> None:
    r"""
    Plot xy scatter animation with given time points

    :param x: x loc. `Array[float, T]`
    :param y: y loc. `Array[float, T]`
    :param t: time array in sec. `Array[float, T]`
    :param size: size of the scatter
    :param step: step run per datapoint
    :param output: output for animation. i.e., \*.gif
    :param kwargs: additional arguments passed to ``FuncAnimation()``
    :return:
    """
    from matplotlib.animation import FuncAnimation

    fig, _ = plt.subplots()

    def foreach_run(frame: int):
        fig.clear()
        ax = fig.add_subplot(111)
        ax.set_xlim(np.min(x), np.max(x))
        ax.set_ylim(np.min(y), np.max(y))

        if step is not None:
            frame *= step

        ax.text(0.02, 0.95, f'Frames = {frame}', transform=ax.transAxes)
        if t is not None:
            ax.text(0.02, 0.85, f'Time = {t[frame]:.2f}', transform=ax.transAxes)

        ax.scatter(x[frame], y[frame], s=size)

    ani = FuncAnimation(fig, foreach_run, frames=len(x), **kwargs)

    try:
        if output is not None:
            ani.save(output)
        else:
            plt.show()

    finally:
        plt.clf()
        plt.close('all')



from pathlib import Path

from neuralib.typing import PathLike

__all__ = ['mkdir_version']


def mkdir_version(root: PathLike,
                  folder_name: str,
                  reuse_latest: bool = False) -> Path:
    """
    Make directory with version number, and auto-accumulate

    :param root: Root directory where the versioned folder will be created.
    :param folder_name: Name of the folder for which versions will be created. Can contain sub-paths.
    :param reuse_latest: Boolean flag indicating whether to reuse the latest version folder if it exists.
    :return: Path of the created or reused version folder.
    """
    root = Path(root)

    if not root.exists():
        root.mkdir(parents=True, exist_ok=True)

    if '/' in folder_name:
        i = folder_name.rindex('/')
        root = root / folder_name[:i]
        folder_name = folder_name[i + 1:]

    output = root / f'{folder_name}_v0'
    end = len(output.name) - 1
    versions = list(root.glob(f'{folder_name}_v*'))
    n_versions = len(versions)

    if output.exists() or n_versions != 0:  # either has v0 or other versions
        v = 1

        if n_versions != 0:
            max_version = max([int(d.name[end:]) for d in versions])
            v = max_version + 1 if not reuse_latest else max_version

        output = root / f'{folder_name}_v{v}'

    output.mkdir(parents=True, exist_ok=True)

    return output



from __future__ import annotations

from pathlib import Path
from typing import Literal, Callable, Sequence

import matplotlib.colors as mcolors
import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.patches import Polygon
from scipy.stats import pearsonr

from neuralib.plot._dotplot import DotPlot
from neuralib.typing import ArrayLikeStr, ArrayLike
from neuralib.typing import PathLike, DataFrame
from neuralib.util.verbose import fprint

__all__ = [
    'dotplot',
    'scatter_histplot',
    'scatter_binx_plot',
    'axvline_histplot',
    'diag_histplot',
    'diag_heatmap',
    'violin_boxplot',
    'grid_subplots',
]


# ========= #
# Dots plot #
# ========= #

def dotplot(xlabel: ArrayLikeStr,
            ylabel: ArrayLikeStr,
            values: np.ndarray,
            *,
            scale: Literal['area', 'radius'] = 'area',
            max_marker_size: float | None = None,
            size_title: str | None = None,
            size_legend_num: int | None = None,
            size_legend_as_int: bool = True,
            with_color: bool = False,
            cmap: mcolors.Colormap = 'Reds',
            colorbar_title: str | None = None,
            norm: mcolors.Normalize | None = None,
            cbar_vmin: float | None = None,
            cbar_vmax: float | None = None,
            figure_title: str | None = None,
            figure_output: PathLike | None = None,
            ax: Axes | None = None,
            **kwargs):
    """
     Plot values as dots, with the option also in colormap

    `Dimension parameters`:

        X = number of x label

        Y = number of y label

    :param xlabel: String arraylike. `ArrayLike[str, X]`
    :param ylabel: String arraylike. `ArrayLike[str, X]`
    :param values: 2D value array. `ArrayLike[str, [X, Y]]`
    :param scale: Dot size representation. {'area', 'radius'}
    :param max_marker_size: Marker size for the max value
    :param size_title: Size_title in the size legend
    :param size_legend_num: Number of legend to be shown
    :param size_legend_as_int: Size legend show only `Int`
    :param with_color: If dot with colormap
    :param cmap: ``Colormap``
    :param colorbar_title: Title of the colorbar
    :param norm: Colorbar Normalize
    :param cbar_vmin: Value min for the colorbar
    :param cbar_vmax: Value nax for the colorbar
    :param figure_title: Figure title
    :param figure_output: Figure save output path
    :param ax: If existing axes ``Axes``
    :param kwargs: additional arguments to ``ax.scatter()``
    """

    dp = DotPlot(xlabel, ylabel, values,
                 scale=scale,
                 max_marker_size=max_marker_size,
                 size_title=size_title,
                 size_legend_num=size_legend_num,
                 size_legend_as_int=size_legend_as_int,
                 with_color=with_color,
                 cmap=cmap,
                 colorbar_title=colorbar_title,
                 norm=norm,
                 cbar_vmin=cbar_vmin,
                 cbar_vmax=cbar_vmax,
                 figure_title=figure_title,
                 figure_output=figure_output,
                 ax=ax)

    dp.plot(**kwargs)


def _ax_size_legend(ax: Axes,
                    value: list[int] | np.ndarray,
                    f: Callable[[float], float]):
    """
    add the labels and legend of the size in scatter plot.

    :param ax: ``Axes``
    :param value: values reflect to size
    :param f: amplified callable
    """
    vsize = np.linspace(0, np.max(value), num=5)

    for s in vsize:
        ax.scatter([], [], s=f(s), c='k', label=str(s))

    handles, labels = plt.gca().get_legend_handles_labels()

    plt.legend(handles[1:], labels[1:], labelspacing=1.2, title="value", borderpad=1,
               frameon=True, framealpha=0.6, edgecolor="k", facecolor="w", loc='right')


# ========== #
# Regression #
# ========== #

def scatter_histplot(x: np.ndarray,
                     y: np.ndarray,
                     bins: int | Sequence[float] | str = 15,
                     *,
                     linear_reg: bool = True,
                     output: Path | None = None,
                     **kwargs):
    """
    plot the linear correlation scatter and histogram between two variables

    `Dimension parameters`:

        N = number of sample points

    :param x: numerical array x. `Array[float, N]`
    :param y: numerical array y. `Array[float, N]`
    :param bins: passed to ``numpy.histogram()``
    :param linear_reg: If show correlation coefficient
    :param output: Figure save output
    :param kwargs: additional args pass through ``ax.set()``
    :return:
    """
    sns.set_theme(style='white', font_scale=1.2)

    g = sns.JointGrid(x=x, y=y, height=5)

    if linear_reg:
        g.plot_joint(sns.regplot, color='grey')
    else:
        g.plot_joint(sns.scatterplot, color='grey')

    g = g.plot_marginals(sns.histplot, kde=False, bins=bins, color='grey')

    ax = g.ax_joint
    ax.set(**kwargs)

    #
    if linear_reg:
        cc = pearsonr(x, y)[0]
        ax.text(0.5, 0.8, f'r = {round(cc, 4)}', fontstyle='italic',
                horizontalalignment='center',
                verticalalignment='center',
                transform=ax.transAxes)

    if output is not None:
        plt.savefig(output)
    else:
        plt.show()


def scatter_binx_plot(ax: Axes,
                      x: np.ndarray,
                      y: np.ndarray,
                      bins: int | Sequence[float] | str = 10,
                      *,
                      order: int = 1,
                      linear_reg: bool = True,
                      bin_func: Literal['median', 'mean'] = 'median',
                      **kwargs):
    """
    Regression to see the relationship between x and y

    `Dimension parameters`:

        N = number of sample points

    :param ax: ``Axes``
    :param x: Numerical array. `Array[float, N]`
    :param y: Numerical array. `Array[float, N]`
    :param bins: passed to ``numpy.histogram()``
    :param order: If order is greater than `1`, use numpy.polyfit to estimate a polynomial regression
    :param linear_reg: Show linear correlation coefficient
    :param bin_func: Literal['median', 'mean']. default is median
    :param kwargs: additional args passed to ``ax.set``
    :return:
    """
    import seaborn as sns

    sns.regplot(
        x=x,
        y=y,
        order=order,
        scatter_kws={'alpha': 0.5,
                     'color': 'grey',
                     's': 8,
                     'edgecolors': 'none'},
        line_kws={'color': 'black'},
        ax=ax
    )
    _hist_line_plot(ax, x, y, bins, bin_func=bin_func)

    ax.set(**kwargs)
    ax.set_aspect(1.0 / ax.get_data_ratio(), adjustable='box')

    if linear_reg:
        cc = pearsonr(x, y)[0]
        ax.set_title(f'r = {round(cc, 4)}', fontstyle='italic')


def _hist_line_plot(ax,
                    x: np.ndarray,
                    y: np.ndarray,
                    bins: int | Sequence[float] | str = 20,
                    bin_func: Literal['median', 'mean'] = 'median'):
    """
    average or pickup the median of the y value in certain bins
    :param ax: ``Axes``
    :param x: `Array[float, N]`
    :param y: `Array[float, N]`
    :param bins:  passed to ``numpy.histogram()``
    :param bin_func
    :return:
    """

    hist, edg = np.histogram(x, weights=y, bins=bins)

    if bin_func == 'mean':
        n = np.histogram(x, bins)[0]  # (B, )
        bin_values = np.divide(hist, n, out=np.zeros_like(hist, dtype=float), where=n != 0)  # avoid true divide
    elif bin_func == 'median':
        bin_values = []
        for i in range(len(hist)):
            bin_data = y[(x >= edg[i]) & (x < edg[i + 1])]
            median_val = np.nanmedian(bin_data)
            bin_values.append(median_val)
    else:
        raise ValueError(f'{bin_func}')

    x = edg[:-1] + np.median(np.diff(edg)) / 2  # alignment
    ax.plot(x, bin_values, 'r--', alpha=0.5)


# ============= #
# Diagonal Plot #
# ============= #


def diag_histplot(x: ArrayLike,
                  y: ArrayLike,
                  bins: int | Sequence[float] | str = 30,
                  *,
                  anchor: Literal['above', 'below'] = 'above',
                  hist_width: float = 0.3,
                  scatter_kws: dict | None = None,
                  polygon_kws: dict | None = None,
                  ax: Axes | None = None) -> None:
    """
    Scatter plot with an overlaid histogram along the diagonal

    `Dimension parameters`:

        N = number of sample points

    :param x: Numerical array. `Array[float, N]`
    :param y: Numerical array. `Array[float, N]`
    :param bins:  Number of bins (or bin specification) for ``np.histogram()``
    :param anchor: Whether the histogram bars extend above or below the main diagonal
    :param hist_width: Maximum length of histogram bars in the rotated space
    :param scatter_kws: Keyword arguments passed to ``ax.scatter()``
    :param polygon_kws: Keyword arguments passed to each ``Polygon()`` patch
    :param ax: ``Axes``
    :return:
    """
    if ax is None:
        _, ax = plt.subplots()

    x = np.array(x)
    y = np.array(y)

    if scatter_kws is None:
        scatter_kws = {'s': 3, 'c': 'black', 'marker': '.', 'edgecolors': 'none'}
    ax.scatter(x, y, **scatter_kws)

    #
    if polygon_kws is None:
        polygon_kws = {'facecolor': 'gray', 'edgecolor': 'none', 'zorder': -1}

    vmin = np.min([x, y])
    vmax = np.max([x, y])

    X = (x - y) / np.sqrt(2) * vmax  # Rotate by +45°
    Y_anchor = vmax / np.sqrt(2)

    # Compute histogram in rotated X coordinate (perpendicular to the anti-diagonal)
    counts, edges = np.histogram(X, bins=bins)
    max_count = counts.max() if counts.max() > 0 else 1

    sign = 1 if anchor == 'above' else -1

    R_inv = np.array([
        [1 / np.sqrt(2), 1 / np.sqrt(2)],
        [-1 / np.sqrt(2), 1 / np.sqrt(2)]
    ])

    for i in range(bins):
        c = counts[i]
        if c == 0:
            continue

        X0, X1 = edges[i], edges[i + 1]  # Bin edges in X (rotated coordinates)
        bar_height = (c / max_count) * hist_width  # Bar height is scaled by the bin count

        # Define the corners of the rectangle in (X, Y) space.
        # The rectangle is anchored at Y_anchor.
        # Its corners: (X0, Y_anchor), (X1, Y_anchor), (X1, Y_anchor + sign*bar_height), (X0, Y_anchor + sign*bar_height)
        corners_rot = np.array([
            [X0, X1, X1, X0],
            [Y_anchor, Y_anchor, Y_anchor + sign * bar_height, Y_anchor + sign * bar_height]
        ])

        # Transform the corners back to original (x,y) coordinates
        corners_xy = R_inv @ corners_rot  # shape (2,4)

        poly = Polygon(corners_xy.T, **polygon_kws)
        ax.add_patch(poly)

    ax.axline((0.5, 0.5), slope=1, color='k', alpha=0.5, lw=1)
    ax.set(xlim=(vmin, vmax), ylim=(vmin, vmax))
    ax.set_aspect('equal', adjustable='box')


def diag_heatmap(x: ArrayLike,
                 y: ArrayLike,
                 cmap: str = 'gist_earth',
                 *,
                 grid_xy: tuple[complex, complex] = (100j, 100j),
                 extent: tuple[float, float, float, float] | None = None,
                 scatter_kws: dict | None = None,
                 imshow_kws: dict | None = None,
                 ax: Axes | None = None) -> None:
    """
    Scatter plot with an overlaid kernel density heatmap

    :param x: Numerical array. `Array[float, N]`
    :param y: Numerical array. `Array[float, N]`
    :param cmap: Colormap used for the heatmap
    :param grid_xy: Grid specification for generating the heatmap
    :param extent: extent for the ``ax.imshow()``
    :param scatter_kws: Keyword arguments passed to ``ax.scatter()``
    :param imshow_kws: Keyword arguments passed to ``ax.imshow()``
    :param ax: ``Axes``
    :return:
    """
    if ax is None:
        _, ax = plt.subplots()

    x = np.array(x)
    y = np.array(y)

    if scatter_kws is None:
        scatter_kws = {'s': 3, 'c': 'gray', 'marker': '.', 'edgecolors': 'none'}
    ax.scatter(x, y, **scatter_kws)

    #
    if extent is None:
        x1, x2 = np.min(x), np.max(x)
        y1, y2 = np.min(y), np.max(y)
    else:
        x1, x2, y1, y2 = extent

    kde = _create_kde_map(x, y, grid_xy, extent=(x1, x2, y1, y2))

    if imshow_kws is None:
        imshow_kws = {}
    ax.imshow(np.rot90(kde), cmap=cmap, extent=(x1, x2, y1, y2), **imshow_kws)

    ax.axline((0.5, 0.5), slope=1, color='k', alpha=0.5, lw=1)
    ax.set_aspect('equal', adjustable='box')


def _create_kde_map(x, y, grid_xy, extent) -> np.ndarray:
    from scipy import stats

    if extent is None:
        x1, x2 = np.min(x), np.max(x)
        y1, y2 = np.min(y), np.max(y)
    else:
        x1, x2, y1, y2 = extent

    xstep, ystep = grid_xy

    X, Y = np.mgrid[x1:x2:xstep, y1:y2:ystep]
    positions = np.vstack([X.ravel(), Y.ravel()])
    values = np.vstack([x, y])
    kernel = stats.gaussian_kde(values)

    return np.reshape(kernel(positions).T, X.shape)


# ====== #
# Others #
# ====== #

def axvline_histplot(ax: Axes,
                     values: np.ndarray,
                     cutoff: float,
                     bins: int = 30,
                     *,
                     mask: np.ndarray | None = None,
                     **kwargs) -> None:
    """
    Plot the histogram with a cutoff value

    :param ax: ``Axes``
    :param values: 1d array. `Array[float, N]`
    :param cutoff: cutoff (threshold) value for the certain value, >= represents pass
    :param bins: passed to ``numpy.histogram()``
    :param mask: mask for value. i.e., cell selection. `Array[bool, N]`
    :param kwargs: passed to ``ax.set``
    """
    if values.ndim != 1:
        raise ValueError('value must be 1d array')

    if mask is not None:
        values = values[mask]

    if cutoff > np.max(values) or cutoff < np.min(values):
        fprint(f'{cutoff} should be within {np.min(values)} and {np.max(values)}', vtype='warning')

    sns.histplot(values, bins=bins, kde=True, color='grey', stat='percent', element='step', ax=ax)
    ax.axvline(cutoff, color='r', linestyle='--', zorder=1)
    ax.set(**kwargs)


def violin_boxplot(ax: Axes,
                   data: DataFrame | dict | list[np.ndarray],
                   x: str | None = None,
                   y: str | None = None,
                   hue: str | None = None,
                   scatter_alpha: float = 0.7,
                   scatter_size: float = 3,
                   output: PathLike | None = None,
                   **kwargs) -> None:
    """
    Plot the data with half violin together with boxes and scatters

    :param ax: ``Axes``
    :param data: Dataset for plotting
    :param x: Names of variables in data or vector data: ``x``
    :param y: Names of variables in data or vector data: ``y``
    :param hue: Names of variables in data or vector data: ``hue``
    :param scatter_alpha: Scatter alpha for the ``sns.stripplot()``
    :param scatter_size: Scatter size for the ``sns.stripplot()``
    :param output: Fig save output path
    :param kwargs: Common args pass through ``sns.violinplot()``, ``sns.boxplot()`` and ``sns.stripplot()``
    :return:
    """
    kwargs = dict(
        ax=ax,
        x=x,
        y=y,
        hue=hue,
        data=data,
        **kwargs
    )

    sns.violinplot(dodge=False, density_norm="width", inner=None, **kwargs)

    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    for violin in ax.collections:
        bbox = violin.get_paths()[0].get_extents()
        x0, y0, width, height = bbox.bounds
        violin.set_clip_path(plt.Rectangle((x0, y0), width / 2, height, transform=ax.transData))

    sns.boxplot(saturation=1,
                showfliers=False,
                width=0.3,
                boxprops={'zorder': 3, 'facecolor': 'none'},
                **kwargs)

    old_len_collections = len(ax.collections)

    sns.stripplot(dodge=False,
                  alpha=scatter_alpha,
                  size=scatter_size,
                  **kwargs)

    for dots in ax.collections[old_len_collections:]:
        dots.set_offsets(dots.get_offsets() + np.array([0.12, 0]))
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)

    if output is not None:
        plt.savefig(output)


def grid_subplots(data: np.ndarray | list[np.ndarray],
                  images_per_row: int,
                  plot_func: Callable | str,
                  *,
                  dtype: Literal['xy', 'img'],
                  hide_axis: bool = True,
                  sharex: bool = False,
                  sharey: bool = False,
                  title: list[str] | None = None,
                  figsize: tuple[int, int] | None = None,
                  output: PathLike | None = None,
                  **kwargs) -> None:
    r"""
    Plots a sequence of subplots in a grid format

    Example for plot xy grid::

        >>> data = np.random.sample((30, 10, 2))
        >>> grid_subplots(data, 5, 'plot', dtype='xy')

    Example for plot img array grid ::

        >>> data = np.random.sample((30, 10, 10))
        >>> grid_subplots(data, 5, 'imshow', dtype='img', cmap='gray')

    :param data: 3D Array containing the data to be plotted. For 'xy' dtype, the shape must be (N, (\*, 2)).
        For 'img' dtype, the shape must be (N, (\*img)). Accepted also list of 2D array different size
    :param images_per_row: Number of images per row in the subplot grid
    :param plot_func: Function or method name to be used for plotting. If a string is provided,
        it should be a valid method name of a matplotlib Axes object
    :param dtype: {'xy', 'img'}. Type of data. 'xy' for (x, y) coordinate data, 'img' for image data
    :param hide_axis: If True, hides the axes of the subplots
    :param sharex: sharex acrross grid plots
    :param sharey: sharey acrross grid plots
    :param title: List of title foreach show in the subplot
    :param figsize: Figure_size pass to ``plt.subplots()``
    :param output: Path to save the plot image. If None, displays the plot.
    :param kwargs: Additional keyword arguments passed to the plotting function ``plot_func``
    :return:
    """
    n_images = len(data)
    n_rows = np.ceil(n_images / images_per_row).astype(int)
    n_cols = min(images_per_row, n_images)

    if figsize is None:
        figsize = (n_cols, n_rows)

    _, ax = plt.subplots(n_rows, n_cols, figsize=figsize, squeeze=False, sharex=sharex, sharey=sharey)

    for i in range(n_rows * n_cols):
        r, c = divmod(i, images_per_row)
        if hide_axis:
            ax[r, c].axis('off')

        #
        if i < n_images:  # check to avoid index error
            if callable(plot_func):
                f = plot_func
                kwargs['ax'] = ax[r, c]
            else:
                f = getattr(ax[r, c], plot_func)

            # dtype
            if dtype == 'xy':
                dat = data[i]
                if dat.shape[1] != 2:
                    raise ValueError(f'invalid {data.size}')
                f(dat[:, 0], dat[:, 1], **kwargs)
            elif dtype == 'img':
                f(data[i], **kwargs)
            else:
                raise ValueError(f'unknown data type: {dtype}')

            # title
            if title is not None:
                ax[r, c].set_title(title[i])

        else:
            ax[r, c].set_visible(False)

    if output is None:
        plt.show()
    else:
        plt.savefig(output)



from __future__ import annotations

import abc
from pathlib import Path
from typing import Literal, get_type_hints, overload, ClassVar, Generic, TypeVar, Any

import h5py
import numpy as np
import polars as pl
from neuralib.util.unstable import unstable
from neuralib.util.verbose import fprint
from numpy.core.numerictypes import issubdtype

__all__ = [
    'H5pyData', 'attr', 'group', 'array', 'table'
]

T = TypeVar('T')
OPEN = Literal['r', 'r+', 'w', 'x', 'a']


@unstable()
class H5pyData:
    READ_ONLY: ClassVar[bool]

    def __init_subclass__(cls, read_only=False, **kwargs):
        cls.READ_ONLY = read_only

    def __init__(self, file: str | Path | h5py.File | h5py.Group,
                 mode: OPEN = 'r'):
        """

        :param file: file path
        """

        if isinstance(file, Path):
            file = str(file)

        if self.READ_ONLY:
            if mode != 'r':
                fprint('read-only wrapper. force read mode')

            mode = 'r'

        if isinstance(file, str):
            file = h5py.File(file, mode)

        self.__file = file

    @property
    def file(self) -> h5py.File | h5py.Group:
        return self.__file

    def __del__(self):
        if isinstance(file := self.__file, h5py.File):
            file.close()

    def __getitem__(self, item):
        return self.__file[item]


def attr(name: str = None):
    return _H5pyAttr(name)


def group(name: str = None):
    return _H5pyGroup(name)


@overload
def array(key: str = None,
          chunks: bool | int | tuple[int, ...] = None,
          maxshape: int | tuple[int, ...] = None,
          compression: int | str = None,
          compression_opts: int | tuple = None,
          scaleoffset: int = None,
          shuffle: bool = None,
          fillvalue: int | float | str = None,
          **kwargs) -> np.ndarray:
    pass


def array(key: str = None, **kwargs) -> np.ndarray:
    return _H5pyArray(key, **kwargs)


def table(key: str = None, backend: Literal['default', 'pytables'] = 'default', **kwargs) -> T:
    if backend == 'default':
        return _H5PyTable_Default(key, **kwargs)
    elif backend == 'pytables':
        return _H5pyTable_PyTable(key, **kwargs)
    else:
        fprint(f'unknown util_h5py.table(backend={backend}). use default.', vtype='warning')
        return _H5PyTable_Default(key, **kwargs)


class _H5pyAttr:
    __slots__ = '__attr', '__type'

    def __init__(self, name: str = None):
        self.__attr = name
        self.__type = None

    def __set_name__(self, owner: type[H5pyData], name: str):
        """called when the descriptor is assigned to a class attribute"""
        if not issubclass(owner, H5pyData):
            raise TypeError('owner type not H5pyDataWrapper')

        if self.__attr is None:
            self.__attr = name
        self.__type = get_type_hints(owner).get(name, None)

    def __get__(self, instance: H5pyData, owner):
        if instance is None:
            return self
        else:
            file: h5py.File = instance.file

            try:
                ret = file.attrs[self.__attr]
            except KeyError as e:
                raise AttributeError(self.__attr) from e

            if self.__type is not None:
                ret = self.__type(ret)
            return ret

    def __set__(self, instance: H5pyData, value):
        try:
            instance.file.attrs[self.__attr] = value
        except OSError as e:
            raise AttributeError(self.__attr) from e

    def __delete__(self, instance: H5pyData):
        try:
            del instance.file.attrs[self.__attr]
        except OSError as e:
            raise AttributeError(self.__attr) from e


class _H5pyGroup:
    __slots__ = '__group', '__type'

    def __init__(self, group: str = None):
        self.__group = group
        self.__type = None

    def __set_name__(self, owner: type[H5pyData], name):
        if not issubclass(owner, H5pyData):
            raise TypeError('owner type not H5pyDataWrapper')

        if self.__group is None:
            self.__group = name

        self.__type = get_type_hints(owner).get(name, None)
        if self.__type is not None and not issubclass(self.__type, H5pyData):
            raise TypeError(f'h5py_group {name} type not H5pyDataWrapper')
        if self.__type is None:
            self.__type = H5pyData

    def __get__(self, instance: H5pyData, owner):
        if instance is None:
            return self
        else:
            file: h5py.File = instance.file

            return self.__type(file.require_group(self.__group))


class _H5pyArray:
    __slots__ = '__key', '__kwargs'

    def __init__(self, key: str = None, **kwargs):
        self.__key = key
        self.__kwargs = kwargs

    def __set_name__(self, owner: type[H5pyData], name):
        if not issubclass(owner, H5pyData):
            raise TypeError('owner type not H5pyDataWrapper')

        if self.__key is None:
            self.__key = name

    def __get__(self, instance: H5pyData, owner):
        if instance is None:
            return self
        else:
            file: h5py.File = instance.file

            try:
                ret = file[self.__key]
            except KeyError:
                pass
            else:
                return _H5pyLazyArray(ret)

            return None

    def __set__(self, instance: H5pyData, value: np.ndarray):
        file: h5py.File = instance.file
        try:
            file[self.__key]
        except KeyError:
            pass
        else:
            del file[self.__key]

        file.create_dataset(self.__key, data=value, **self.__kwargs)

    def __delete__(self, instance: H5pyData):
        file: h5py.File = instance.file
        del file[self.__key]


class _H5pyLazyArray:
    def __init__(self, data: h5py.Dataset):
        self.__data = data

    @property
    def ndim(self):
        return self.__data.ndim

    @property
    def dtype(self):
        return self.__data.dtype

    @property
    def shape(self):
        return self.__data.shape

    def __array__(self, *args, **kwargs):
        return np.asarray(self.__data, *args, **kwargs)

    def __getitem__(self, item):
        return np.asarray(self.__data[item])


class _H5pyTable(Generic[T], metaclass=abc.ABCMeta):
    __slots__ = '__key', '_type', '_kwargs'

    def __init__(self, key: str = None, **kwargs):
        self.__key = key
        self._type = None
        self._kwargs = kwargs

    def __set_name__(self, owner, name):
        if not issubclass(owner, H5pyData):
            raise TypeError('owner type not H5pyDataWrapper')

        if self.__key is None:
            self.__key = name

        self._type = get_type_hints(owner).get(name, None)

    @abc.abstractmethod
    def _get_table(self, table: h5py.Group) -> T:
        pass

    @abc.abstractmethod
    def _set_table(self, group: h5py.Group, table: T):
        pass

    def __get__(self, instance: H5pyData, owner) -> T:
        if instance is None:
            return self
        else:
            file: h5py.File = instance.file

            try:
                ret = file[self.__key]
            except KeyError:
                pass
            else:
                return self._get_table(ret)

            return None

    def __set__(self, instance: H5pyData, value: T):
        file: h5py.File = instance.file

        try:
            ret = file[self.__key]
            if not isinstance(ret, h5py.Group):
                del file[self.__key]
                raise KeyError
        except KeyError:
            ret = file.create_group(self.__key)

        self._set_table(ret, value)

    def __delete__(self, instance: H5pyData):
        file: h5py.File = instance.file
        del file[self.__key]


class _H5PyTable_Default(_H5pyTable[pl.DataFrame]):

    def _get_table(self, table: h5py.Group) -> pl.DataFrame:
        import polars.datatypes as pty

        print(f'{dir(table)=}')
        attrs = table['schema'].attrs
        content = table['table']

        schema = {
            name: getattr(pty, attrs[name])
            for name in attrs
        }

        data = {
            name: np.asarray(content[name])
            for name in schema
        }

        return pl.DataFrame(data=data, schema_overrides=schema)

    def _set_table(self, group: h5py.Group, table: pl.DataFrame):
        for name in table.schema:
            if (dtype := table.schema[name]).is_nested():
                raise RuntimeError(f'Do not support nested data type : {name} <{dtype}>')

        try:
            schema = group.create_group('schema')
        except ValueError:
            del group['schema']
            schema = group.create_group('schema')

        try:
            content = group.create_group('table')
        except ValueError:
            del group['table']
            content = group.create_group('table')

        for name in table.schema:
            schema.attrs[name] = str(table.schema[name])

        for name in table.columns:
            content.create_dataset(name, data=table[name].to_numpy())


class _H5pyTable_PyTable(_H5pyTable[Any]):
    def __init__(self, key: str = None, **kwargs):
        super().__init__(key, **kwargs)

    def _get_table(self, table: h5py.Group):
        import pandas as pd
        df = pd.read_hdf(table, **self._kwargs)
        if issubdtype(self._type, pd.DataFrame):
            return df

        import polars as pl
        df = pl.from_pandas(df)
        if issubdtype(self._type, pl.DataFrame):
            return df

        try:
            from neuralib.util.util_polars import DataFrameWrapper
            if issubdtype(self._type, DataFrameWrapper):
                return self._type(df)
        except ImportError:
            # in ts-dev branch
            pass

        return df

    def _set_table(self, group: h5py.Group, table):
        raise RuntimeError('unsupported now')



from __future__ import annotations

import typing
from typing import Literal

import attrs
import numpy as np
from typing_extensions import Self

from neuralib.imaging import ObjectiveFov

__all__ = ['CellularCoordinates']


@typing.final
@attrs.define
class CellularCoordinates:
    """Container with coordinates information (in mm) for each ROIs

    `Dimension parameters`:

        N: number of neurons

        P: number of plane types

    """

    neuron_idx: np.ndarray
    """neuron index. `Array[float, N]`"""

    ap: np.ndarray
    """anterior posterior coordinates (default in mm). `Array[float, N]`"""

    ml: np.ndarray
    """medial lateral coordinates (default in mm). `Array[float, N]`"""

    plane_index: int | np.ndarray | None = attrs.field(default=None, kw_only=True)
    """optic plane index. i.e., used for depth analysis. If array type, `Array[int, P]`"""

    source_plane_index: np.ndarray | None = attrs.field(default=None, kw_only=True)
    """neuron's corresponding image plane. `Array[float, N]`"""

    value: np.ndarray | None = attrs.field(default=None, kw_only=True)
    """metric (i.e., used in topographical analysis). `Array[float, N]` """

    unit: Literal['mm', 'um'] = attrs.field(default='mm', kw_only=True, validator=attrs.validators.in_(('mm', 'um')))
    """unit of the coordinates. `{'mm', 'um'}` """

    @source_plane_index.validator
    def check(self, attribute, value):
        if self.source_plane_index is not None:
            if isinstance(self.plane_index, int):
                if np.any(value != self.plane_index):
                    raise RuntimeError('invalid source plane index')
            elif isinstance(self.plane_index, np.ndarray):
                if np.any([v not in self.plane_index for v in value]):
                    raise RuntimeError('invalid source plane index')
            else:
                raise TypeError(f'{attribute} type error')

    def to_um(self) -> Self:
        """unit from mm to um"""
        if self.unit == 'um':
            raise RuntimeError('unit already in um')

        return attrs.evolve(
            self,
            ap=self.ap * 1000,
            ml=self.ml * 1000,
            unit='um'
        )

    def with_value(self, value: np.ndarray) -> Self:
        """
        assign ``value`` for ``CellularCoordinates``

        :param value: value array for the corresponding coordinates. `Array[float, N]`
        :return:
        """

        assert len(value) == len(self.ap) == len(self.ml)
        return attrs.evolve(self, value=value)

    def with_selection(self, mask: np.ndarray) -> Self:
        """masking for cell selection

        :param mask: numpy bool array. `Array[bool, N]`
        """
        return attrs.evolve(
            self,
            neuron_idx=self.neuron_idx[mask],
            ap=self.ap[mask],
            ml=self.ml[mask],
            value=None if self.value is None else self.value[mask]
        )

    def in_relative_bregma(self, fov: ObjectiveFov) -> Self:
        """TODO check
        register cellular coordinates to IBL coordinates space

        :param fov: :class:`~neuralib.imaging.fov.ObjectiveFov`
        """
        if fov.rotation_angle_ap != 0:
            raise NotImplementedError('')

        fov = fov.to_um()

        zelf = self.to_um() if self.unit != 'um' else self

        if fov.rotation_angle_ml != 0:
            rx, ry = self.ml_rotate(zelf.ml, zelf.ap, fov.rotation_angle_ml)
            return attrs.evolve(self, ap=fov.am[1] - ry, ml=fov.am[0] - rx)
        else:
            return attrs.evolve(self, ap=fov.am[1] - zelf.ap, ml=fov.am[0] - zelf.ml)

    # ========================= #
    # Handle Objective rotation #
    # ========================= #

    # TODO need further testing. the drifted in this method increased across depth
    @staticmethod
    def ml_rotate(x: np.ndarray, y: np.ndarray, deg: float) -> tuple[np.ndarray, np.ndarray]:
        """
        calculate the values for x axis rotation

        :param x: coordinate x. `Array[float, N]`
        :param y: coordinate Y. `Array[float, N]`
        :param deg: rotation degree
        :return: value to be subtracted. dx (N,) and dy (N,)
        """
        angle = np.radians(deg)
        dx = x * np.cos(angle) - y * np.sin(angle)
        dy = x * np.sin(angle) + y * np.cos(angle)

        return dx, dy

    @staticmethod
    def ap_rotate():
        # TODO
        pass



from __future__ import annotations

import re
from collections.abc import Collection
from dataclasses import is_dataclass
from pathlib import Path
from typing import TypeVar, overload, Literal, Any

from neuralib.typing import PathLike
from neuralib.typing.func import is_namedtuple
from neuralib.util.verbose import fprint

__all__ = ['uglob',
           'filter_matched',
           'joinn',
           'ensure_dir',
           'keys_with_value',
           'cls_hasattr']


def uglob(directory: PathLike,
          pattern: str,
          is_dir: bool = False) -> Path:
    """
    Use glob pattern to find the unique file in the directory.

    :param directory: Directory
    :param pattern: Glob pattern
    :param is_dir: Is the pattern point to a directory?
    :return: The unique path
    :raise FileNotFoundError: the unique path is not existed.
    :raise NotADirectoryError: *directory* is not a directory
    :raise RuntimeError: more than one path are found.
    """
    directory = Path(directory)

    if not directory.exists():
        raise FileNotFoundError(f'{directory} not exit')

    if not directory.is_dir():
        raise NotADirectoryError(f'{directory} is not a directory')

    f = list(directory.glob(pattern))

    if is_dir:
        f = [ff for ff in f if ff.is_dir()]
    else:
        f = [ff for ff in f if not ff.is_dir()]

    if len(f) == 0:
        t = 'directory' if is_dir else 'file'
        raise FileNotFoundError(f'{directory} not have {t} with the pattern: {pattern}')
    elif len(f) == 1:
        return f[0]
    else:
        f.sort()
        t = 'directories' if is_dir else 'files'
        raise RuntimeError(f'multiple {t} were found in {directory} with the pattern {pattern} >>> {f}')


def filter_matched(pattern: str, strings: list[str]) -> list[str]:
    """
    Filter a list of string element that match the pattern.

    :param pattern: Regular expression
    :param strings: List of strings to find the pattern
    :return:
    """
    return list(filter(re.compile(pattern).match, strings))


def ensure_dir(p: PathLike, verbose: bool = True) -> Path:
    """
    Ensure the path is a directory. Create if it is not exist.

    :param p: path to be checked
    :param verbose: print verbose message
    """
    p = Path(p)

    if not p.exists():
        p.mkdir(parents=True)

        if verbose:
            fprint(f'create dir {p}', vtype='io')

    if not p.is_dir():
        raise NotADirectoryError(f'not a dir: {p}')

    return p


def joinn(sep: str, *part: str | None) -> str:
    """join non-None str with sep."""
    return sep.join([str(it) for it in part if it is not None])


# ============================== #

KT = TypeVar('KT')
VT = TypeVar('VT')


@overload
def keys_with_value(dy: dict[KT, Any | Collection[VT]],
                    value: VT,
                    to_item: Literal[False] = False) -> list[KT]:
    pass


@overload
def keys_with_value(dy: dict[KT, VT | Collection[VT]],
                    value: VT,
                    to_item: Literal[True] = True) -> KT:
    pass


def keys_with_value(dy, value, to_item=False):
    """
    Get keys from a dict that are associated with the value.

    Supports value types: str, int, float (with tolerance), and any collection types.

    :param dy: The value to match against the dictionary values
    :param value: The value to match against the dictionary value
    :param to_item: List of item to item if unique
    :return: A list of keys whose values match the provided value
    """
    matching_keys = []

    def _float_eq(v1, v2, tol=1e-9) -> bool:
        return abs(v1 - v2) < tol

    for key, val in dy.items():
        if isinstance(val, float) and isinstance(value, float):
            if _float_eq(val, value):
                matching_keys.append(key)
        elif isinstance(val, (str, int, Collection)) or is_namedtuple(val) or is_dataclass(val):
            if val == value:
                matching_keys.append(key)

    if to_item:
        if len(matching_keys) == 0:
            raise KeyError(f'{value} does not match any keys')
        elif len(matching_keys) == 1:
            return matching_keys[0]
        else:
            raise ValueError(f'multiple match keys: {matching_keys}')
    else:
        return matching_keys


def cls_hasattr(cls: type, attr: str) -> bool:
    """
    Check if attributes in class

    :param cls: The class to check for the attribute.
    :param attr: The name of the attribute to look for within the class and its hierarchy.
    :return: True if the class or any of its parent classes has the specified attribute, False otherwise.
    """
    if attr in getattr(cls, '__annotations__', {}):
        return True

    for c in cls.mro()[1:]:  # Skip the first class as it's already checked
        if attr in getattr(c, '__annotations__', {}):
            return True

    return False



from __future__ import annotations

import datetime
from typing import overload, Any, TypeVar, Union

from .annotation import *
from .expr import *
from .expr import use_table_first, SqlRemovePlaceHolder
from .literal import UPDATE_POLICY
from .stat import *
from .table import *

__all__ = [
    'create_table',
    'insert_into',
    'replace_into',
    'select_from',
    'update',
    'delete_from'
]

T = TypeVar('T')
S = TypeVar('S')


@overload
def select_from(table: type[T], *, distinct: bool = False) -> SqlSelectStat[T]:
    """
    >>> select_from(Table) # SELECT * FROM Table
    """
    pass


@overload
def select_from(table: SqlCteExpr, *, distinct: bool = False) -> SqlSelectStat[tuple]:
    pass


@overload
def select_from(*field, distinct: bool = False,
                from_table: Union[str, type, SqlAlias, SqlSelectStat] = None) -> SqlSelectStat[tuple]:
    """
    >>> select_from('a', 'b') # SELECT a, b FROM Table
    """
    pass


def select_from(*args, distinct: bool = False,
                from_table: Union[str, type, SqlAlias, SqlSelectStat] = None) -> SqlSelectStat:
    """
    ``SELECT``: https://www.sqlite.org/lang_select.html

    Select all fields from a table

    >>> select_from(A).build() # doctest: SKIP
    SELECT * FROM A
    >>> select_from(A).fetchall() # doctest: SKIP
    [A(...), A(...), ...]

    Select subset of fields from A

    >>> select_from(A.a, A.b).build() # doctest: SKIP
    SELECT A.a, A.b FROM A
    >>> select_from(A.a, A.b).fetchall() # doctest: SKIP
    [('a', 1), ('b', 2), ...]

    With a literal value

    >>> select_from(A.a, 0).build() # doctest: SKIP
    SELECT A.a, 0 FROM A
    >>> select_from(A.a, 0).fetchall() # doctest: SKIP
    [('a', 0), ('b', 0), ...]

    With SQL functions

    >>> select_from(A.a, count()).build() # doctest: SKIP
    SELECT A.a, COUNT(*) FROM A

    Use table alias

    >>> a = alias(A, 'a') # doctest: SKIP
    >>> select_from(a.a).build() # doctest: SKIP
    SELECT a.a from A a

    join other tables

    >>> select_from(A.a, B.b).join(A.c == B.c).build() # doctest: SKIP
    SELECT A.a, B.b FROM A JOIN B ON A.c = B.c

    **features supporting**

    * `SELECT DISTINCT`
    * `FROM`
    * `WHERE`
    * `GROUP BY`
    * `HAVING`
    * `WINDOW`
    * compound-operator: `UNION [ALL]`, `INTERSECT` and `EXCEPT`
    * `ORDER BY`
    * `LIMIT [OFFSET]`

    **features not supporting**

    * `WITH [RECURSIVE]`
    * `SELECT ALL`
    * `VALUES`

    :param args:
    :param distinct:
    :param from_table:
    :return:
    """
    pre_stat = ['SELECT']
    if distinct:
        pre_stat.append('DISTINCT')

    if len(args) == 1 and isinstance(table := args[0], type):
        self = SqlSelectStat(table)
        self._involved.append(table)
        self.add(pre_stat)
        self.add('*')
    elif len(args) == 1 and isinstance(table := args[0], SqlCteExpr):
        self = SqlSelectStat(None)
        self.add(table)
        self.add(pre_stat)
        self.add('*')

    else:

        table, fields = select_from_fields(*args)
        if from_table is not None:
            table = from_table

        if table is None:
            raise RuntimeError('need to provide from_table')

        if isinstance(table, type):
            self = SqlSelectStat(None)
            self._involved.append(table)
        elif isinstance(table, SqlAlias) and isinstance(table._value, type):
            self = SqlSelectStat(None)
            self._involved.append(table)
        elif isinstance(table, SqlCteExpr):
            self = SqlSelectStat(None)
            self.add(table)
        else:
            self = SqlSelectStat(None)

        self.add(pre_stat)

        for i, field in enumerate(fields):
            if i > 0:
                self.add(',')

            if isinstance(field, SqlField):
                self.add(f'{field.table_name}.{field.name}')
            elif isinstance(field, SqlAlias) and isinstance(_field := field._value, SqlField):
                self.add([f'{_field.table_name}.{_field.name}', 'AS', repr(field._name)])
            elif isinstance(field, SqlAlias) and isinstance(expr := field._value, SqlExpr):
                self.add(expr)
                self.add(['AS', repr(field._name)])
            elif isinstance(field, SqlAliasField):
                self.add(field)
            elif isinstance(field, SqlFuncOper):
                self.add(field)
            elif isinstance(field, SqlLiteral):
                self.add(field)
            elif isinstance(field, SqlExpr):
                self.add(field)
            else:
                self.drop()
                raise TypeError('SELECT ' + repr(field))

    with self:
        self.add('FROM')
        if isinstance(table, str):
            self.add(table)
        elif isinstance(table, type):
            self.add(table_name(table))
        elif isinstance(table, SqlStat):
            self.add('(')
            self.add(table)
            self.add(')')
        elif isinstance(table, SqlAlias) and isinstance(_table := table._value, type):
            self.add([table_name(_table), table._name])
        elif isinstance(table, SqlAlias) and isinstance(subq := table._value, SqlSubQuery):
            self.add('(')
            self.add(subq.stat)
            self.add([')', 'AS', repr(table._name)])
        elif isinstance(table, SqlCteExpr):
            self.add(table._name)
        else:
            raise TypeError('FROM ' + repr(table))

    return self


def select_from_fields(*args) -> tuple[Union[type, SqlAlias, None], list[SqlExpr]]:
    if len(args) == 0:
        raise RuntimeError('empty field')

    table = None
    fields = []
    for arg in args:
        if isinstance(arg, (int, float, bool, str)):
            fields.append(SqlLiteral(repr(arg)))

        elif isinstance(arg, Field):
            if table is None:
                table = arg.table

            fields.append(SqlField(arg))

        elif isinstance(arg, SqlField):
            if table is None:
                table = arg.table

            fields.append(arg)

        elif isinstance(arg, SqlAlias) and isinstance(arg._value, SqlField):
            if table is None:
                table = arg._value.table

            fields.append(arg)

        elif isinstance(arg, SqlAliasField) and isinstance(field_table:=arg.table, type):
            if table is None:
                table = SqlAlias(field_table, arg.name)

            fields.append(arg)

        elif isinstance(arg, SqlAliasField) and isinstance(cte := arg.table, SqlCteExpr):
            if table is None:
                table = cte

            fields.append(arg)

        elif isinstance(arg, SqlExpr):
            if table is None:
                table =  use_table_first(arg)

            fields.append(arg)

        elif isinstance(arg, property):
            if (expr_table := getattr(arg.fget, '_sql_owner', None)) is None:
                raise RuntimeError(f'{arg} not a property from a Table')

            if table is None:
                table = expr_table

            if isinstance(expr := arg.fget(expr_table), SqlExpr):
                fields.append(expr)
            else:
                raise RuntimeError(f'{arg} does not return a sql expression.')
        else:
            raise TypeError(repr(arg))

    return table, fields


@overload
def insert_into(table: type[T], *, policy: UPDATE_POLICY = None, named=False) -> SqlInsertStat[T]:
    pass


@overload
def insert_into(*field, policy: UPDATE_POLICY = None, named=False) -> SqlInsertStat[T]:
    pass


def insert_into(*args, policy: UPDATE_POLICY = None, named=False) -> SqlInsertStat[T]:
    """
    ``INSERT``: https://www.sqlite.org/lang_insert.html

    insert values

    >>> insert_into(A, policy='REPLACE').build() # doctest: SKIP
    INSERT OR REPLACE INTO A VALUES (?)
    >>> insert_into(A, policy='REPLACE').submit([A(1), A(2)]) # doctest: SKIP

    insert values with field overwrite

    >>> insert_into(A, policy='REPLACE').values(a='1').build() # doctest: SKIP
    INSERT OR REPLACE INTO A VALUES (1)

    insert values from a table

    >>> insert_into(A, policy='IGNORE').select_from(B).build() # doctest: SKIP
    INSERT OR IGNORE INTO A
    SELECT * FROM B

    **features supporting**

    * `INSERT [OR ...]`
    * `VALUES`
    * `DEFAULT VALUES`
    * `SELECT`
    * upsert clause
    * returning clause

    **features not supporting**

    * `WITH [RECURSIVE]`

    :param table:
    :param policy:
    :param named:
    :return:
    """
    if len(args) == 1 and isinstance(args[0], type):
        self = SqlInsertStat((table := args[0]), named=named)
    else:
        table, fields = select_from_fields(*args)
        if table is None:
            raise RuntimeError()

        for i, field in enumerate(list(fields)):
            if isinstance(field, SqlField):
                if field.table != table:
                    raise RuntimeError(f'field {field.table_name}.{field.name} not belong to {table.__name__}')
                fields[i] = field.name
            else:
                raise TypeError()

        self = SqlInsertStat(table, fields, named=named)

    with self:
        self.add('INSERT')
        if policy is not None:
            self.add(['OR', policy.upper()])
        self.add(['INTO', table_name(table)])
        if self._fields is not None:
            self.add('(')
            for i, field in enumerate(self._fields):
                if i > 0:
                    self.add(',')
                self.add(field)
            self.add(')')
        return self


@overload
def replace_into(table: type[T], *, named=False) -> SqlInsertStat[T]:
    pass


@overload
def replace_into(*field: Any, named=False) -> SqlInsertStat[T]:
    pass


def replace_into(*args, named=False) -> SqlInsertStat[T]:
    return insert_into(*args, policy='REPLACE', named=named)


def update(table: type[T], *args: Union[bool, SqlCompareOper], **kwargs) -> SqlUpdateStat[T]:
    """
    ``UPDATE``: https://www.sqlite.org/lang_update.html

    >>> update(A, A.a==1).where(A.b==2).build() # doctest: SKIP
    UPDATE A SET A.a = 1 WHERE A.b = 2

     **features supporting**

    * `UPDATE [OR ...]`
    * `SET COLUMN = EXPR`
    * `FROM`
    * `WHERE`
    * `ON CONFLICT (COLUMNS) SET (COLUMNS) = EXPR`
    * returning clause

    **features not supporting**

    * `WITH [RECURSIVE]`
    * (qualified table name) `INDEXED BY`
    * (qualified table name) `NOT INDEXED`

    :param table:
    :param args:
    :param kwargs:
    :return:
    """
    with SqlUpdateStat(table) as self:
        self.add(['UPDATE', table_name(table), 'SET'])

        if len(args):
            for arg in args:
                self.add(SqlCompareOper.as_set_expr(arg))
                self.add(',')

        for term, value in kwargs.items():
            table_field(table, term)
            self.add(SqlLiteral(term) == value)
            self.add(',')
        self._stat.pop()

        return self


def delete_from(table: type[T]) -> SqlDeleteStat[T]:
    """
    ``DELETE``: https://www.sqlite.org/lang_delete.html

    >>> delete_from(A).where(A.b > 2).build()  # doctest: SKIP
    DELETE FROM A WHERE A.b > 2

    **features supporting**

    * `DELETE FROM`
    * `WHERE`
    * `ORDER BY`
    * `LIMIT [OFFSET]`
    * returning clause

    **features not supporting**

    * `WITH [RECURSIVE]`
    * (qualified table name) `INDEXED BY`
    * (qualified table name) `NOT INDEXED`

    :param table:
    :return:
    """
    with SqlDeleteStat(table) as self:
        self.add(['DELETE', 'FROM', table_name(table)])
        return self


def create_table(table: type[T], *, if_not_exists=True) -> SqlStat[T]:
    """
    ``CREATE``: https://www.sqlite.org/lang_createtable.html

    >>> @named_tuple_table_class # doctest: SKIP
    ... class A(NamedTuple):
    ...     a: int
    >>> create_table(A) # doctest: SKIP
    CREATE TABLE IF NOT EXISTS A (a INT NOT NULL)

    **features supporting**

    * `IF NOT EXISTS`
    * column constraint `NOT NULL`
    * column constraint `PRIMARY KEY`
    * column constraint `UNIQUE`
    * column constraint `CHECK`
    * column constraint `DEFAULT value`
    * table constraint `PRIMARY KEY`
    * table constraint `UNIQUE`
    * table constraint `CHECK`
    * table constraint `FOREIGN KEY`

    **features not supporting**

    * `CREATE TEMP|TEMPORARY`
    * `CREATE TEMP`
    * `AS SELECT`
    * column constraint `CONSTRAINT`
    * column constraint `NOT NULL ON CONFLICT`
    * column constraint `DEFAULT (EXPR)`
    * column constraint `COLLATE`
    * column constraint `REFERENCES`
    * column constraint `[GENERATED ALWAYS] AS`
    * table constraint `CONSTRAINT`
    * `WITHOUT ROWID`
    * `STRICT`

    :param table:
    :return:
    """
    with SqlStat(table) as self:
        self.add(['CREATE', 'TABLE'])
        if if_not_exists:
            self.add('IF NOT EXISTS')
        self.add(table_name(table))
        self.add('(')

        n_primary_key = len(primary_keys := table_primary_fields(table))

        for i, field in enumerate(table_fields(table)):
            if n_primary_key == 1 and field.is_primary:
                column_def(self, field, field.get_primary())
            else:
                column_def(self, field)
            self.add(',')

        if n_primary_key > 1:
            self.add(['PRIMARY KEY', '(', ' , '.join([it.name for it in primary_keys]), ')'])
            for it in primary_keys:
                if (conflict := it.get_primary().conflict) is not None:
                    self.add(['ON CONFLICT', conflict.upper()])
                    break
            self.add(',')

        for unique in table_unique_fields(table):
            if len(unique.fields) > 1:
                self.add(['UNIQUE', '(', ' , '.join(unique.fields), ')'])
                if (conflict := unique.conflict) is not None:
                    self.add(['ON CONFLICT', conflict.upper()])
                self.add(',')

        for foreign_key in table_foreign_fields(table):
            foreign_constraint(self, foreign_key)
            self.add(',')

        if (check := table_check_field(table, None)) is not None:
            check_constraint(self, check)
            self.add(',')

        self._stat.pop()
        self.add(')')

        return self


def column_def(self: SqlStat, field: Field, primary: PRIMARY = None):
    self.add(f"[{field.name}]")

    if field.sql_type == Any:
        pass
    elif field.sql_type == int:
        self.add('INTEGER')
    elif field.sql_type == float:
        self.add('FLOAT')
    elif field.sql_type == bool:
        self.add('BOOLEAN')
    elif field.sql_type == bytes:
        self.add('BLOB')
    elif field.sql_type == str:
        self.add('TEXT')
    elif field.sql_type == datetime.time:
        self.add('DATETIME')
    elif field.sql_type == datetime.date:
        self.add('DATETIME')
    elif field.sql_type == datetime.datetime:
        self.add('DATETIME')
    else:
        raise RuntimeError(f'field type {field.sql_type}')

    if field.not_null:
        if not field.has_default or field.f_value is not None:
            self.add('NOT NULL')

    if primary is not None:
        self.add(['PRIMARY', 'KEY'])
        if primary.order is not None:
            self.add(primary.order.upper())
        if primary.conflict is not None:
            self.add(['ON CONFLICT', primary.conflict.upper()])
        if primary.auto_increment:
            self.add('AUTOINCREMENT')

    elif (unique := field.get_unique()) is not None:
        self.add('UNIQUE')
        if unique.conflict is not None:
            self.add(['ON CONFLICT', unique.conflict.upper()])

    from .table import missing
    if field.f_value is missing:
        pass
    elif field.f_value is None:
        self.add('DEFAULT NULL')
    elif field.f_value == CURRENT_DATE:
        self.add('DEFAULT CURRENT_DATE')
    elif field.f_value == CURRENT_TIME:
        self.add('DEFAULT CURRENT_TIME')
    elif field.f_value == CURRENT_TIMESTAMP:
        self.add('DEFAULT CURRENT_TIMESTAMP')
    else:
        self.add(f'DEFAULT {repr(field.f_value)}')

    if (check := table_check_field(field.table, field.name)) is not None:
        check_constraint(self, check)


def foreign_constraint(self: SqlStat, foreign: ForeignConstraint):
    self.add(['FOREIGN KEY'])
    self.add('(')
    self.add(' , '.join(foreign.fields))
    self.add(')')
    self.add('REFERENCES')
    self.add(table_name(foreign.foreign_table))
    self.add('(')
    self.add(' , '.join(foreign.foreign_fields))
    self.add(')')
    if (policy := foreign.on_update) is not None:
        self.add(['ON UPDATE', policy])
    if (policy := foreign.on_delete) is not None:
        self.add(['ON DELETE', policy])


def check_constraint(self: SqlStat, check: CheckConstraint):
    self._deparameter = True
    self.add(['CHECK', '('])
    self.add(SqlRemovePlaceHolder(check.expression))
    self.add(')')
    self._deparameter = False



from typing import Literal

from typing_extensions import TypeAlias

__all__ = ['Area', 'Source', 'Channel', 'MergeLevel',
           'HEMISPHERE_TYPE']

Area: TypeAlias = str
"""Brain area"""

Source: TypeAlias = str
"""Name of the injection alias"""

Channel: TypeAlias = str
"""Fluorescence channel"""

MergeLevel: TypeAlias = int
"""Hierarchical tree merge level"""

#
HEMISPHERE_TYPE = Literal['ipsi', 'contra', 'both']
"""brain hemisphere"""



"""
Annotation-based argparse
=========================

:author:
    Ta-Shun Su

This module provide a way to integrate python ``argparse`` module into class attribute
and annotation type, that allow options have type information, and allow parser combination
easily.

class Argument and function argument
------------------------------------

A simple option class that contains several options carried by their attributes.

>>> class ExampleOptions:
...     ANIMAL: str = argument('--ANIMAL')
...     EXP_DATE: str = argument('--EXP_DATE')
...     OUTPUT_DIR: str = argument()

In ``ExampleOptions``, ``ANIMAL`` is an attribute with type annotation ``str``. it has
a class variable :class:`~neuralib.argp.core.Argument` (:func:`~neuralib.argp.core.argument()` return) which contains the
arguments of ``ArgumentParser.add_argument``. For now, we have an optional
argument ``--ANIMAL`` which accept one argument. ``EXP_DATE`` as is another optional argument.
And ``OUTPUT_DIR`` are a potional argument because it doesn't have dashed options.

>>> opt = parse_args(ExampleOptions())
... print(opt.ANIMAL)

After class declared, you can use :meth:`~neuralib.argp.core.parse_args()` to parse cli arguments. This
function will create an ``ArgumentParser`` and find out all argument attributes.
Then set the attributes from parsed result.

Commandline usage
~~~~~~~~~~~~~~~~~

In bash, you can call this option class ::

    python -m module.path --ANIMAL name --EXP_DATE date output

:func:`~neuralib.argp.core.print_help()` does the similar things but print the help document to the stdout.

>>> print_help(opt)

Or use ``-h`` options ::

    python -m module.path -h


Annotation type infering
~~~~~~~~~~~~~~~~~~~~~~~~

In general, you can think :func:`~neuralib.argp.core.argument` just a delegate funcion that passes the arguments
to the ``ArgumentParser.add_argument``. However, this function will try to
infer missing arguments based on the annotation type when creating the ``ArgumentParser``.
For now, this module is not powerful to handle all possible case. there are support type
(Please see :meth:`~neuralib.argp.core.Argument.complete_kwargs()` for detailed):

1. ``bool``: infer parameter ``action`` to ``store_true``, ``default`` to ``False``.
2. ``Literal[...]`` : infer parameter ``choices``.
3. ``Optional[T]``: infer parameter ``type`` to ``T`` if it is callable. If it is ``Literal``, apply 2.
4. ``callable(...)`` with signature ``Callable[[str], T]``: infer parameter ``type`` to ``T``
5. parameter ``dest`` always use the attribute name.

Additionally, :class:`~neuralib.argp.core.Argument` also provide a parameter ``group`` to reduce the complexity of
create subgrouping parser.

Option class compose
~~~~~~~~~~~~~~~~~~~~

Option class can be composed by inherition. Child option class can also change the value from parent's
argument. As well as disable it (by replacing a value)

>>> class MoreOptions(ExampleOptions):
...     # additional optional option
...     verbose: bool = argument('-v', '--verbose')
...     # change default value
...     ANIMAL: str = as_argument(ExampleOptions.animal).with_options(default='YW00')
...     # and disable an option
...     OUTPUT_DIR: str = 'output' # just replace with a value

Change options name is more complicate, because you might want to add more name, remove some name,
or rename some name. :meth:`~neuralib.argp.core.Argument.with_options()` allow you to do that:

>>> class ChangeExample(ExampleOptions):
...     # replace option name: --animal
...     ANIMAL: str = as_argument(ExampleOptions.animal).with_options('--animal')
...     # add more option name: -A, --ANIMAL
...     ANIMAL: str = as_argument(ExampleOptions.animal).with_options(..., '-A')
...     # rename option: --animal
...     ANIMAL: str = as_argument(ExampleOptions.animal).with_options({
...         '--ANIMAL': '--animal'
...     })

Utility function for option class
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

``ExampleOptions`` doesn't declated neither ``__str__`` nor ``__repr__``, so it is not convenient to debug.
funciont :func:`~neuralib.argp.core.as_dict()` provide a way to take argument attribute's value into a dictionary.

>>> as_dict(opt)
{'ANIMAL': ..., 'EXP_DATE': ..., 'OUTPUT_DIR': ...}

Option class is not restricted into only one use case. It works like a normal class.

class AbstractParser
--------------------

This class provide a main like class that has more control on ``argparse.ArgumentParser``
creation.

>>> class ExampleParser(AbstractParser, ExampleOptions):
...     DESCRIPTION = 'Example parser'
...     def run(self):
...         ...
>>> if __name__ == '__main__':
...     ExampleParser().main()

Subcommands
-----------

This module isn't fully support sub-command feature, but only provide a simple way for specific case:

>>> parse_command_args(
...     description='top level parser',
...     parsers=dict(ep=ExampleParser)
... )

In bash::

    python -m module.path ep --ANIMAL name --EXP_DATE date output

"""
from ._type import *
from .core import *
